from typing import Any, Dict, Iterator, List, Optional, Union  # noqa: F401
from IPython.display import HTML, Image  # noqa: F401

from itertools import chain
from pathlib import Path
import os
import re

from pandas import DataFrame, Series
import lunchbox.tools as lbt
import networkx
import numpy as np
import pandas as pd

import rolling_pin.tools as rpt
# ------------------------------------------------------------------------------

'''
Contains the RepoETL class, which is used for converted python repository module
dependencies into a directed graph.
'''


class RepoETL():
    '''
    RepoETL is a class for extracting 1st order dependencies of modules within a
    given repository. This information is stored internally as a DataFrame and
    can be rendered as networkx, pydot or SVG graphs.
    '''
    def __init__(
        self,
        root,
        include_regex=r'.*\.py$',
        exclude_regex=r'(__init__|test_|_test|mock_)\.py$',
    ):
        # type: (Union[str, Path], str, str) -> None
        r'''
        Construct RepoETL instance.

        Args:
            root (str or Path): Full path to repository root directory.
            include_regex (str, optional): Files to be included in recursive
                directy search. Default: '.*\.py$'.
            exclude_regex (str, optional): Files to be excluded in recursive
                directy search. Default: '(__init__|test_|_test|mock_)\.py$'.

        Raises:
            ValueError: If include or exclude regex does not end in '\.py$'.
        '''
        self._root = root  # type: Union[str, Path]
        self._data = self._get_data(root, include_regex, exclude_regex)  # type: DataFrame

    @staticmethod
    def _get_imports(fullpath):
        # type: (Union[str, Path]) -> List[str]
        '''
        Get's import statements from a given python module.

        Args:
            fullpath (str or Path): Path to python module.

        Returns:
            list(str): List of imported modules.
        '''
        with open(fullpath) as f:
            data = f.readlines()  # type: Union[List, Iterator]
        data = map(lambda x: x.strip('\n'), data)
        data = filter(lambda x: re.search('^import|^from', x), data)
        data = map(lambda x: re.sub('from (.*?) .*', '\\1', x), data)
        data = map(lambda x: re.sub(' as .*', '', x), data)
        data = map(lambda x: re.sub(' *#.*', '', x), data)
        data = map(lambda x: re.sub('import ', '', x), data)
        data = filter(lambda x: not lbt.is_standard_module(x), data)
        return list(data)

    @staticmethod
    def _get_data(
        root,
        include_regex=r'.*\.py$',
        exclude_regex=r'(__init__|_test)\.py$',
    ):
        # type: (Union[str, Path], str, str) -> DataFrame
        r'''
        Recursively aggregates and filters all the files found with a given
        directory into a DataFrame. Data is used to create directed graphs.

        DataFrame has these columns:

            * node_name    - name of node
            * node_type    - type of node, can be [module, subpackage, library]
            * x            - node's x coordinate
            * y            - node's y coordinate
            * dependencies - parent nodes
            * subpackages  - parent nodes of type subpackage
            * fullpath     - fullpath to the module a node represents

        Args:
            root (str or Path): Root directory to be searched.
            include_regex (str, optional): Files to be included in recursive
                directy search. Default: '.*\.py$'.
            exclude_regex (str, optional): Files to be excluded in recursive
                directy search. Default: '(__init__|_test)\.py$'.

        Raises:
            ValueError: If include or exclude regex does not end in '\.py$'.
            FileNotFoundError: If no files are found after filtering.

        Returns:
            DataFrame: DataFrame of file information.
        '''
        root = Path(root).as_posix()
        files = rpt.list_all_files(root)  # type: Union[Iterator, List]
        if include_regex != '':
            if not include_regex.endswith(r'\.py$'):
                msg = f"Invalid include_regex: '{include_regex}'. "
                msg += r"Does not end in '.py$'."
                raise ValueError(msg)

            files = filter(
                lambda x: re.search(include_regex, x.absolute().as_posix()),
                files
            )
        if exclude_regex != '':
            files = filter(
                lambda x: not re.search(exclude_regex, x.absolute().as_posix()),
                files
            )

        files = list(files)
        if len(files) == 0:
            msg = f'No files found after filters in directory: {root}.'
            raise FileNotFoundError(msg)

        # buid DataFrame of nodes and imported dependencies
        data = DataFrame()
        data['fullpath'] = files
        data.fullpath = data.fullpath.apply(lambda x: x.absolute().as_posix())

        data['node_name'] = data.fullpath\
            .apply(lambda x: re.sub(root, '', x))\
            .apply(lambda x: re.sub(r'\.py$', '', x))\
            .apply(lambda x: re.sub('^/', '', x))\
            .apply(lambda x: re.sub('/', '.', x))

        data['subpackages'] = data.node_name\
            .apply(lambda x: rpt.get_parent_fields(x, '.')).apply(lbt.get_ordered_unique)
        data.subpackages = data.subpackages\
            .apply(lambda x: list(filter(lambda y: y != '', x)))

        data['dependencies'] = data.fullpath\
            .apply(RepoETL._get_imports).apply(lbt.get_ordered_unique)
        data.dependencies += data.node_name\
            .apply(lambda x: ['.'.join(x.split('.')[:-1])])
        data.dependencies = data.dependencies\
            .apply(lambda x: list(filter(lambda y: y != '', x)))

        data['node_type'] = 'module'

        # add subpackages as nodes
        pkgs = set(chain(*data.subpackages.tolist()))  # type: Any
        pkgs = pkgs.difference(data.node_name.tolist())
        pkgs = sorted(list(pkgs))
        pkgs = Series(pkgs)\
            .apply(
                lambda x: dict(
                    node_name=x,
                    node_type='subpackage',
                    dependencies=rpt.get_parent_fields(x, '.'),
                    subpackages=rpt.get_parent_fields(x, '.'),
                )).tolist()
        pkgs = DataFrame(pkgs)
        data = pd.concat([data, pkgs], ignore_index=True, sort=True)

        # add library dependencies as nodes
        libs = set(chain(*data.dependencies.tolist()))  # type: Any
        libs = libs.difference(data.node_name.tolist())
        libs = sorted(list(libs))
        libs = Series(libs)\
            .apply(
                lambda x: dict(
                    node_name=x,
                    node_type='library',
                    dependencies=[],
                    subpackages=[],
                )).tolist()
        libs = DataFrame(libs)
        data = pd.concat([data, libs], ignore_index=True, sort=True)

        data.drop_duplicates('node_name', inplace=True)
        data.reset_index(drop=True, inplace=True)

        # define node coordinates
        data['x'] = 0
        data['y'] = 0
        data = RepoETL._calculate_coordinates(data)
        data = RepoETL._anneal_coordinate(data, 'x', 'y')
        data = RepoETL._center_coordinate(data, 'x', 'y')

        data.sort_values('fullpath', inplace=True)
        data.reset_index(drop=True, inplace=True)

        cols = [
            'node_name',
            'node_type',
            'x',
            'y',
            'dependencies',
            'subpackages',
            'fullpath',
        ]
        data = data[cols]
        return data

    @staticmethod
    def _calculate_coordinates(data):
        # type: (DataFrame) -> DataFrame
        '''
        Calculate inital x, y coordinates for each node in given DataFrame.
        Node are startified by type along the y axis.

        Args:
            DataFrame: DataFrame of nodes.

        Returns:
            DataFrame: DataFrame with x and y coordinate columns.
        '''
        # set initial node coordinates
        data['y'] = 0
        for item in ['module', 'subpackage', 'library']:
            mask = data.node_type == item
            n = data[mask].shape[0]

            index = data[mask].index
            data.loc[index, 'x'] = list(range(n))

            # move non-library nodes down the y axis according to how nested
            # they are
            if item != 'library':
                data.loc[index, 'y'] = data.loc[index, 'node_name']\
                    .apply(lambda x: len(x.split('.')))

        # move all module nodes beneath supackage nodes on the y axis
        max_ = data[data.node_type == 'subpackage'].y.max()
        index = data[data.node_type == 'module'].index
        data.loc[index, 'y'] += max_
        data.loc[index, 'y'] += data.loc[index, 'subpackages'].apply(len)

        # reverse y axis
        max_ = data.y.max()
        data.y = -1 * data.y + max_

        return data

    @staticmethod
    def _anneal_coordinate(data, anneal_axis='x', pin_axis='y', iterations=10):
        # type: (DataFrame, str, str, int) -> DataFrame
        '''
        Iteratively align nodes in the anneal axis according to the mean
        position of their connected nodes. Node anneal coordinates are rectified
        at the end of each iteration according to a pin axis, so that they do
        not overlap. This mean that they are sorted at each level of the pin
        axis.

        Args:
            data (DataFrame): DataFrame with x column.
            anneal_axis (str, optional): Coordinate column to be annealed.
                Default: 'x'.
            pin_axis (str, optional): Coordinate column to be held constant.
                Default: 'y'.
            iterations (int, optional): Number of times to update x coordinates.
                Default: 10.

        Returns:
            DataFrame: DataFrame with annealed anneal axis coordinates.
        '''
        x = anneal_axis
        y = pin_axis
        data[x] = data[x].astype(float)
        data[y] = data[y].astype(float)
        for iteration in range(iterations):
            # create directed graph from data
            graph = RepoETL._to_networkx_graph(data)

            # reverse connectivity every other iteration
            if iteration % 2 == 0:
                graph = graph.reverse()

            # get mean coordinate of each node in directed graph
            for name in graph.nodes:
                tree = networkx.bfs_tree(graph, name)
                mu = np.mean([graph.nodes[n][x] for n in tree])
                graph.nodes[name][x] = mu

            # update data coordinate column
            for node in graph.nodes:
                mask = data[data.node_name == node].index
                data.loc[mask, x] = graph.nodes[node][x]

            # rectify data coordinate column, so that no two nodes overlap
            data.sort_values(x, inplace=True)
            for yi in data[y].unique():
                mask = data[data[y] == yi].index
                values = data.loc[mask, x].tolist()
                values = list(range(len(values)))
                data.loc[mask, x] = values

        return data

    @staticmethod
    def _center_coordinate(data, center_axis='x', pin_axis='y'):
        # (DataFrame, str, str) -> DataFrame
        '''
        Sorted center_axis coordinates at each level of the pin axis.

        Args:
            data (DataFrame): DataFrame with x column.
            anneal_column (str, optional): Coordinate column to be annealed.
                Default: 'x'.
            pin_axis (str, optional): Coordinate column to be held constant.
                Default: 'y'.
            iterations (int, optional): Number of times to update x coordinates.
                Default: 10.

        Returns:
            DataFrame: DataFrame with centered center axis coordinates.
        '''
        x = center_axis
        y = pin_axis
        max_ = data[x].max()
        for yi in data[y].unique():
            mask = data[data[y] == yi].index
            l_max = data.loc[mask, x].max()
            delta = max_ - l_max
            data.loc[mask, x] += (delta / 2)
        return data

    @staticmethod
    def _to_networkx_graph(data, escape_chars=False):
        # (DataFrame, bool) -> networkx.DiGraph
        '''
        Converts given DataFrame into networkx directed graph.

        Args:
            data (DataFrame): DataFrame of nodes.
            escape_chars (bool, optional): Escape special characters. Used to
                avoid dot file errors. Default: False.

        Returns:
            networkx.DiGraph: Graph of nodes.
        '''
        # escape periods for dot file interpolation
        if escape_chars:
            data = data.copy()
            data.node_name = data.node_name \
                .fillna('') \
                .apply(lambda x: re.sub(r'\.', '\\.', x))
            data.dependencies = data.dependencies \
                .apply(lambda x: [re.sub(r'\.', '\\.', y) for y in x])

        graph = networkx.DiGraph()
        data.apply(
            lambda x: graph.add_node(
                x.node_name,
                **{k: getattr(x, k) for k in x.index}
            ),
            axis=1
        )

        data.apply(
            lambda x: [graph.add_edge(p, x.node_name) for p in x.dependencies],
            axis=1
        )
        return graph

    def to_networkx_graph(self):
        # () -> networkx.DiGraph
        '''
        Converts internal data into networkx directed graph.

        Returns:
            networkx.DiGraph: Graph of nodes.
        '''
        return RepoETL._to_networkx_graph(self._data)

    def to_dot_graph(self, orient='tb', orthogonal_edges=False, color_scheme=None):
        # (str, bool, Optional[Dict[str, str]]) -> pydot.Dot
        '''
        Converts internal data into pydot graph.

        Args:
            orient (str, optional): Graph layout orientation. Default: tb.
                Options include:

                * tb - top to bottom
                * bt - bottom to top
                * lr - left to right
                * rl - right to left
            orthogonal_edges (bool, optional): Whether graph edges should have
                non-right angles. Default: False.
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.tools.COLOR_SCHEME

        Raises:
            ValueError: If orient is invalid.

        Returns:
            pydot.Dot: Dot graph of nodes.
        '''
        orient = orient.lower()
        orientations = ['tb', 'bt', 'lr', 'rl']
        if orient not in orientations:
            msg = f'Invalid orient value. {orient} not in {orientations}.'
            raise ValueError(msg)

        # set color scheme of graph
        if color_scheme is None:
            color_scheme = rpt.COLOR_SCHEME

        # create dot graph
        graph = self._to_networkx_graph(self._data, escape_chars=True)
        dot = networkx.drawing.nx_pydot.to_pydot(graph)

        # set layout orientation
        dot.set_rankdir(orient.upper())

        # set graph background color
        dot.set_bgcolor(color_scheme['background'])

        # set edge draw type
        if orthogonal_edges:
            dot.set_splines('ortho')

        # set draw parameters for each node in graph
        for node in dot.get_nodes():
            # set node shape, color and font attributes
            node.set_shape('rect')
            node.set_style('filled')
            node.set_color(color_scheme['node'])
            node.set_fillcolor(color_scheme['node'])
            node.set_fontname('Courier')

            nx_node = re.sub('"', '', node.get_name())
            nx_node = graph.nodes[nx_node]

            # if networkx node has no attributes skip it
            # this should not ever occur but might
            if nx_node == {}:
                continue  # pragma: no cover

            # set node x, y coordinates
            node.set_pos(f"{nx_node['x']},{nx_node['y']}!")

            # vary node font color by noe type
            if nx_node['node_type'] == 'library':
                node.set_fontcolor(color_scheme['node_library_font'])
            elif nx_node['node_type'] == 'subpackage':
                node.set_fontcolor(color_scheme['node_subpackage_font'])
            else:
                node.set_fontcolor(color_scheme['node_module_font'])

        # set draw parameters for each edge in graph
        for edge in dot.get_edges():
            # get networkx source node of edge
            nx_node = dot.get_node(edge.get_source())
            nx_node = nx_node[0].get_name()
            nx_node = re.sub('"', '', nx_node)
            nx_node = graph.nodes[nx_node]

            # if networkx source node has no attributes skip it
            # this should not ever occur but might
            if nx_node == {}:
                continue  # pragma: no cover

            # vary edge color by its source node type
            if nx_node['node_type'] == 'library':
                edge.set_color(color_scheme['edge_library'])
            elif nx_node['node_type'] == 'subpackage':
                edge.set_color(color_scheme['edge_subpackage'])
            else:
                # this line is actually covered by pytest doesn't think so
                edge.set_color(color_scheme['edge_module'])  # pragma: no cover

        return dot

    def to_dataframe(self):
        # type: () -> DataFrame
        '''
        Retruns:
            DataFrame: DataFrame of nodes representing repo modules.
        '''
        return self._data.copy()

    def to_html(
        self,
        layout='dot',
        orthogonal_edges=False,
        color_scheme=None,
        as_png=False
    ):
        # type: (str, bool, Optional[Dict[str, str]], bool) -> Union[HTML, Image]
        '''
        For use in inline rendering of graph data in Jupyter Lab.

        Args:
            layout (str, optional): Graph layout style.
                Options include: circo, dot, fdp, neato, sfdp, twopi.
                Default: dot.
            orthogonal_edges (bool, optional): Whether graph edges should have
                non-right angles. Default: False.
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.tools.COLOR_SCHEME
            as_png (bool, optional): Display graph as a PNG image instead of
                SVG. Useful for display on Github. Default: False.

        Returns:
            IPython.display.HTML: HTML object for inline display.
        '''
        if color_scheme is None:
            color_scheme = rpt.COLOR_SCHEME

        dot = self.to_dot_graph(
            orthogonal_edges=orthogonal_edges,
            color_scheme=color_scheme,
        )
        return rpt.dot_to_html(dot, layout=layout, as_png=as_png)

    def write(
        self,
        fullpath,
        layout='dot',
        orient='tb',
        orthogonal_edges=False,
        color_scheme=None
    ):
        # type: (Union[str, Path], str, str, bool, Optional[Dict[str, str]]) -> RepoETL
        '''
        Writes internal data to a given filepath.
        Formats supported: svg, dot, png, json.

        Args:
            fulllpath (str or Path): File to be written to.
            layout (str, optional): Graph layout style.
                Options include: circo, dot, fdp, neato, sfdp, twopi. Default: dot.
            orient (str, optional): Graph layout orientation. Default: tb.
                Options include:

                * tb - top to bottom
                * bt - bottom to top
                * lr - left to right
                * rl - right to left
            orthogonal_edges (bool, optional): Whether graph edges should have
                non-right angles. Default: False.
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.tools.COLOR_SCHEME

        Raises:
            ValueError: If invalid file extension given.

        Returns:
            RepoETL: Self.
        '''
        if isinstance(fullpath, Path):
            fullpath = fullpath.absolute().as_posix()

        _, ext = os.path.splitext(fullpath)
        ext = re.sub(r'^\.', '', ext)
        if re.search('^json$', ext, re.I):
            self._data.to_json(fullpath, orient='records')
            return self

        if color_scheme is None:
            color_scheme = rpt.COLOR_SCHEME

        graph = self.to_dot_graph(
            orient=orient,
            orthogonal_edges=orthogonal_edges,
            color_scheme=color_scheme,
        )
        try:
            rpt.write_dot_graph(graph, fullpath, layout=layout,)
        except ValueError:
            msg = f'Invalid extension found: {ext}. '
            msg += 'Valid extensions include: svg, dot, png, json.'
            raise ValueError(msg)
        return self



import subprocess

import click

from rolling_pin.conform_etl import ConformETL
from rolling_pin.radon_etl import RadonETL
from rolling_pin.repo_etl import RepoETL
from rolling_pin.toml_etl import TomlETL
# ------------------------------------------------------------------------------

'''
Command line interface to rolling-pin library
'''


@click.group()
def main():
    pass


@main.command()
@click.argument('source', type=str, nargs=1)
@click.option(
    '--groups',
    type=str,
    nargs=1,
    default='all',
    help="Comma separated list of groups to be conformed. Default: 'all'"
)
@click.option(
    '--dryrun',
    is_flag=True,
    help="Print out conform table instead of run conform."
)
def conform(source, groups, dryrun):
    # type: (str, str, bool) -> None
    '''
    Copies source files to target filepaths according to given conform file.

    \b
    Arguments:
        SOURCE - conform YAML filepath
    '''
    etl = ConformETL.from_yaml(source)
    if dryrun:
        print(etl)
    else:
        etl.conform(groups=groups.split(','))


@main.command()
@click.argument('source', type=str, nargs=1)
@click.argument('target', type=str, nargs=1)
@click.option(
    '--include',
    type=str,
    nargs=1,
    default=r'.*\.py$',
    help=r"include files that match this regex pattern. default: '.*\.py$'"
)
@click.option(
    '--exclude',
    type=str,
    nargs=1,
    default='test|mock',
    help="exclude files that match this regex pattern. default: 'test|mock'"
)
@click.option(
    '--orient', type=str, nargs=1, default='lr',
    help='graph orientation. default: lr.',
)
def graph(source, target, include, exclude, orient):
    # type: (str, str, str, str, str) -> None
    '''
    Generate a dependency graph of a source repository and write it to a given
    filepath

    \b
    Arguments:
        SOURCE - repository path
        TARGET - target filepath
    '''
    include_ = '' if include is None else include
    exclude_ = '' if exclude is None else exclude
    RepoETL(source, include_, exclude_).write(target, orient=orient)


@main.command()
@click.argument('source', type=str, nargs=1)
@click.argument('target', type=str, nargs=1)
def plot(source, target):
    # type: (str, str) -> None
    '''
    Write radon metrics plots of given repository to given filepath

    \b
    Arguments:
        SOURCE - repository path
        TARGET - plot filepath
    '''
    RadonETL(source).write_plots(target)


@main.command()
@click.argument('source', type=str, nargs=1)
@click.argument('target', type=str, nargs=1)
def table(source, target):
    # type: (str, str) -> None
    '''
    Write radon metrics tables of given repository to given directory

    \b
    Arguments:
        SOURCE - repository path
        TARGET - table directory
    '''
    RadonETL(source).write_tables(target)


@main.command()
@click.argument('source', type=str, nargs=1)
@click.option(
    '--edit',
    type=str,
    nargs=1,
    multiple=True,
    help='''Replace key\'s value with given value. TEXT is "=" separated key
value pair in TOML format''',
)
@click.option(
    '--delete',
    type=str,
    nargs=1,
    multiple=True,
    help='Delete keys that match this regular expression',
)
@click.option(
    '--search',
    type=str,
    nargs=1,
    help='Search for keys that match this regular expression',
)
@click.option(
    '--target',
    type=str,
    nargs=1,
    help='Target filepath to write to',
)
def toml(source, edit, delete, search, target):
    # type: (str, tuple[str], tuple[str], str, str) -> None
    '''
    Generate a copy of a given TOML file with given edits indicated by flags.
    Flags are evalauted in the following order: edit, delete, search.
    Flags may be arbitrarily combined.
    Edit and delete flags may appear multiple times.

    \b
    Arguments:
        SOURCE - TOML filepath

    \b
    EXAMPLES
        EXAMPLE-FILE------------------------------------------------------------
            >>>cat example.toml
            [root]
            a = 1
            b = 2
    \b
            [root.foo.bar]
            x = "y"
    \b
            [world]
            hello = true

    \b
        EDIT-FLAG---------------------------------------------------------------
            >>>rolling-pin toml foobar.toml --edit 'root.a=999'
            [root]
            a = 999
            b = 2...
    \b
            --------------------------------------------------------------------
            >>>rolling-pin toml foobar.toml \\
                   --edit 'root.a=[1, 2]'   \\
                   --edit 'root.b="xxx"'
            [root]
            a = [
                1,
                2,
            ]
            b = "xxx"...
    \b
            --------------------------------------------------------------------
            >>>rolling-pin toml foobar.toml --edit 'root.foo.bar="baz"'
            ...
            hello = true
    \b
            [root.foo]
            bar = "baz"...

    \b
        DELETE-FLAG-------------------------------------------------------------
            >>>rolling-pin toml foobar.toml \\
                   --delete 'root.foo.bar'  \\
                   --delete 'root.a'
            [root]
            b = 2
    \b
            [world]
            hello = true

    \b
        SEARCH-FLAG-------------------------------------------------------------
            >>>rolling-pin toml foobar.toml --search 'root.foo|world'
            [world]
            hello = true
    \b
            [root.foo.bar]
            x = "y"
    '''
    etl = TomlETL.from_toml(source)
    for e in edit:
        etl = etl.edit(e)
    for d in delete:
        etl = etl.delete(d)
    if search is not None:
        etl = etl.search(search)
    if target is not None:
        etl.write(target)
    else:
        print(etl.to_string())


@main.command()
def bash_completion():
    '''
    BASH completion code to be written to a _rolling-pin completion file.
    '''
    cmd = '_ROLLING_PIN_COMPLETE=bash_source rolling-pin'
    result = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    result.wait()
    click.echo(result.stdout.read())


@main.command()
def zsh_completion():
    '''
    ZSH completion code to be written to a _rolling-pin completion file.
    '''
    cmd = '_ROLLING_PIN_COMPLETE=zsh_source rolling-pin'
    result = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    result.wait()
    click.echo(result.stdout.read())


if __name__ == '__main__':
    main()



from typing import Any, Dict, List, Union  # noqa: F401

import json
import os
import re
from pathlib import Path

from pandas import DataFrame
from radon.cli import CCHarvester, HCHarvester, MIHarvester, RawHarvester
from radon.cli import Config
import numpy as np
import pandas as pd
import plotly.express as px
import radon.complexity

from rolling_pin.blob_etl import BlobETL
import rolling_pin.tools as rpt
# ------------------------------------------------------------------------------

'''
Contain the RadonETL class, which is used for generating a radon report on the
code wthin a given directory.
'''


class RadonETL():
    '''
    Conforms all four radon reports (raw metrics, Halstead, maintainability and
    cyclomatic complexity) into a single DataFrame that can then be plotted.
    '''
    def __init__(self, fullpath):
        # type: (Union[str, Path]) -> None
        '''
        Constructs a RadonETL instance.

        Args:
            fullpath (str or Path): Python file or directory of python files.
        '''
        self._report = RadonETL._get_radon_report(fullpath)
    # --------------------------------------------------------------------------

    @property
    def report(self):
        # type: () -> Dict
        '''
        dict: Dictionary of all radon metrics.
        '''
        return self._report

    @property
    def data(self):
        # type: () -> DataFrame
        '''
        DataFrame: DataFrame of all radon metrics.
        '''
        return self._get_radon_data()

    @property
    def raw_metrics(self):
        # type: () -> DataFrame
        '''
        DataFrame: DataFrame of radon raw metrics.
        '''
        return self._get_raw_metrics_dataframe(self._report)

    @property
    def maintainability_index(self):
        # type: () -> DataFrame
        '''
        DataFrame: DataFrame of radon maintainability index metrics.
        '''
        return self._get_maintainability_index_dataframe(self._report)

    @property
    def cyclomatic_complexity_metrics(self):
        # type: () -> DataFrame
        '''
        DataFrame: DataFrame of radon cyclomatic complexity metrics.
        '''
        return self._get_cyclomatic_complexity_dataframe(self._report)

    @property
    def halstead_metrics(self):
        # type: () -> DataFrame
        '''
        DataFrame: DataFrame of radon Halstead metrics.
        '''
        return self._get_halstead_dataframe(self._report)
    # --------------------------------------------------------------------------

    def _get_radon_data(self):
        # type: () -> DataFrame
        '''
        Constructs a DataFrame representing all the radon reports generated for
        a given python file or directory containing python files.

        Returns:
            DataFrame: Radon report DataFrame.
        '''
        hal = self.halstead_metrics
        cc = self.cyclomatic_complexity_metrics
        raw = self.raw_metrics
        mi = self.maintainability_index

        data = hal.merge(cc, how='outer', on=['fullpath', 'name'])
        data['object_type'] = data.object_type_x
        mask = data.object_type_x.apply(pd.isnull)
        mask = data[mask].index
        data.loc[mask, 'object_type'] = data.loc[mask, 'object_type_y']
        del data['object_type_x']
        del data['object_type_y']

        module = raw.merge(mi, on='fullpath')

        cols = set(module.columns.tolist())  # type: Any
        cols = cols.difference(data.columns.tolist())
        cols = list(cols)
        for col in cols:
            data[col] = np.nan

        mask = data.object_type == 'module'
        for i, row in data[mask].iterrows():
            for col in cols:
                val = module[module.fullpath == row.fullpath][col].item()
                data.loc[i, col] = val

        cols = [
            'fullpath', 'name', 'class_name', 'object_type', 'blank', 'bugs',
            'calculated_length', 'code', 'column_offset', 'comment',
            'cyclomatic_complexity', 'cyclomatic_rank', 'difficulty', 'effort',
            'h1', 'h2', 'length', 'logical_code', 'maintainability_index',
            'maintainability_rank', 'multiline_comment', 'n1', 'n2',
            'single_comment', 'source_code', 'start_line', 'stop_line', 'time',
            'vocabulary', 'volume',
        ]
        data = data[cols]

        return data
    # --------------------------------------------------------------------------

    @staticmethod
    def _get_radon_report(fullpath):
        # type: (Union[str, Path]) -> Dict[str, Any]
        '''
        Gets all 4 report from radon and aggregates them into a single blob
        object.

        Args:
            fullpath (str or Path): Python file or directory of python files.

        Returns:
            dict: Radon report blob.
        '''
        fullpath_ = [Path(fullpath).absolute().as_posix()]  # type: List[str]
        output = []  # type: Any

        config = Config(
            min='A',
            max='F',
            exclude=None,
            ignore=None,
            show_complexity=False,
            average=False,
            total_average=False,
            order=getattr(
                radon.complexity, 'SCORE', getattr(radon.complexity, 'SCORE')
            ),
            no_assert=False,
            show_closures=False,
        )
        output.append(CCHarvester(fullpath_, config).as_json())

        config = Config(
            exclude=None,
            ignore=None,
            summary=False,
        )
        output.append(RawHarvester(fullpath_, config).as_json())

        config = Config(
            min='A',
            max='C',
            exclude=None,
            ignore=None,
            multi=True,
            show=False,
            sort=False,
        )
        output.append(MIHarvester(fullpath_, config).as_json())

        config = Config(
            exclude=None,
            ignore=None,
            by_function=False,
        )
        output.append(HCHarvester(fullpath_, config).as_json())

        output = list(map(json.loads, output))
        keys = [
            'cyclomatic_complexity', 'raw_metrics', 'maintainability_index',
            'halstead_metrics',
        ]
        output = dict(zip(keys, output))
        return output

    @staticmethod
    def _get_raw_metrics_dataframe(report):
        # type: (Dict) -> DataFrame
        '''
        Converts radon raw metrics report into a pandas DataFrame.

        Args:
            report (dict): Radon report blob.

        Returns:
            DataFrame: Raw metrics DataFrame.
        '''
        raw = report['raw_metrics']
        fullpaths = list(raw.keys())
        path_lut = {k: f'<list_{i}>' for i, k in enumerate(fullpaths)}
        fullpath_fields = {x: {'fullpath': x} for x in fullpaths}

        #   loc = Lines of Code (total lines) - sloc + blanks + multi + single_comments
        #   lloc = Logical Lines of Code
        #   comments = Comments lines
        #   multi = Multi-line strings (assumed to be docstrings)
        #   blank = Blank lines (or whitespace-only lines)
        #   single_comments = Single-line comments or docstrings
        name_lut = dict(
            blank='blank',
            comments='comment',
            lloc='logical_code',
            loc='code',
            multi='multiline_comment',
            single_comments='single_comment',
            sloc='source_code',
            fullpath='fullpath',
        )
        data = BlobETL(raw, '#')\
            .update(fullpath_fields) \
            .set_field(0, lambda x: path_lut[x])\
            .set_field(1, lambda x: name_lut[x])\
            .to_dict()  # type: Union[Dict, DataFrame]

        data = DataFrame(data)
        data.sort_values('fullpath', inplace=True)
        data.reset_index(drop=True, inplace=True)
        cols = [
            'fullpath', 'blank', 'code', 'comment', 'logical_code',
            'multiline_comment', 'single_comment', 'source_code',
        ]
        data = data[cols]

        return data

    @staticmethod
    def _get_maintainability_index_dataframe(report):
        # type: (Dict) -> DataFrame
        '''
        Converts radon maintainability index report into a pandas DataFrame.

        Args:
            report (dict): Radon report blob.

        Returns:
            DataFrame: Maintainability DataFrame.
        '''
        mi = report['maintainability_index']
        fullpaths = list(mi.keys())
        path_lut = {k: f'<list_{i}>' for i, k in enumerate(fullpaths)}
        fullpath_fields = {x: {'fullpath': x} for x in fullpaths}
        name_lut = dict(
            mi='maintainability_index',
            rank='maintainability_rank',
            fullpath='fullpath',
        )
        data = None  # type: Any
        data = BlobETL(mi, '#')\
            .update(fullpath_fields) \
            .set_field(0, lambda x: path_lut[x])\
            .set_field(1, lambda x: name_lut[x])\
            .to_dict()

        data = DataFrame(data)
        data.sort_values('fullpath', inplace=True)
        data.reset_index(drop=True, inplace=True)
        cols = ['fullpath', 'maintainability_index', 'maintainability_rank']
        data = data[cols]

        # convert rank to integer
        rank_lut = {k: i for i, k in enumerate('ABCDEF')}
        data['maintainability_rank'] = data['maintainability_rank']\
            .apply(lambda x: rank_lut[x])

        return data

    @staticmethod
    def _get_cyclomatic_complexity_dataframe(report):
        # type: (Dict) -> DataFrame
        '''
        Converts radon cyclomatic complexity report into a pandas DataFrame.

        Args:
            report (dict): Radon report blob.

        Returns:
            DataFrame: Cyclomatic complexity DataFrame.
        '''
        filters = [
            [4, 6, 'method_closure',
                '^[^#]+#<list_[0-9]+>#methods#<list_[0-9]+>#closures#<list_[0-9]+>#[^#]+$'],
            [3, 4, 'closure', '^[^#]+#<list_[0-9]+>#closures#<list_[0-9]+>#[^#]+$'],
            [3, 4, 'method', '^[^#]+#<list_[0-9]+>#methods#<list_[0-9]+>#[^#]+$'],
            [2, 2, None, '^[^#]+#<list_[0-9]+>#[^#]+$'],
        ]  # type: Any

        cc = report['cyclomatic_complexity']
        rows = []
        for i, j, type_, regex in filters:
            temp = BlobETL(cc, '#').query(regex)  # type: DataFrame
            if len(temp.to_flat_dict().keys()) > 0:
                temp = temp.to_dataframe(i)
                item = temp\
                    .apply(lambda x: dict(zip(x[j], x['value'])), axis=1)\
                    .tolist()
                item = DataFrame(item)
                item['fullpath'] = temp[0]
                if type_ is not None:
                    item.type = type_
                rows.append(item)
        data = pd.concat(rows, ignore_index=True, sort=False)

        cols = [
            'fullpath', 'name', 'classname', 'type', 'complexity', 'rank',
            'lineno', 'endline', 'col_offset'
        ]
        data = data[cols]
        lut = {
            'fullpath': 'fullpath',
            'name': 'name',
            'classname': 'class_name',
            'type': 'object_type',
            'complexity': 'cyclomatic_complexity',
            'rank': 'cyclomatic_rank',
            'lineno': 'start_line',
            'endline': 'stop_line',
            'col_offset': 'column_offset',
        }
        data.drop_duplicates(inplace=True)
        data.rename(mapper=lambda x: lut[x], axis=1, inplace=True)
        data.reset_index(drop=True, inplace=True)

        # convert rank to integer
        rank_lut = {k: i for i, k in enumerate('ABCDEF')}
        data['cyclomatic_rank'] = data['cyclomatic_rank']\
            .apply(lambda x: rank_lut[x])

        return data

    @staticmethod
    def _get_halstead_dataframe(report):
        # type: (Dict) -> DataFrame
        '''
        Converts radon Halstead report into a pandas DataFrame.

        Args:
            report (dict): Radon report blob.

        Returns:
            DataFrame: Halstead DataFrame.
        '''
        hal = report['halstead_metrics']
        keys = [
            'h1', 'h2', 'n1', 'n2', 'vocabulary', 'length', 'calculated_length',
            'volume', 'difficulty', 'effort', 'time', 'bugs',
        ]
        data = BlobETL(hal, '#').query('function|closure').to_dataframe(3)
        data['fullpath'] = data[0]
        data['object_type'] = data[1].apply(lambda x: re.sub('s$', '', x))
        data['name'] = data.value.apply(lambda x: x[0])

        score = data.value.apply(lambda x: dict(zip(keys, x[1:]))).tolist()
        score = DataFrame(score)
        data = data.join(score)

        total = BlobETL(hal, '#').query('total').to_dataframe()
        total['fullpath'] = total[0]
        total = total.groupby('fullpath', as_index=False)\
            .agg(lambda x: dict(zip(keys, x)))
        score = total.value.tolist()
        score = DataFrame(score)
        total = total.join(score)
        total['object_type'] = 'module'
        total['name'] = total.fullpath\
            .apply(lambda x: os.path.splitext((Path(x).name))[0])
        data = pd.concat([data, total], ignore_index=True, sort=False)

        cols = ['fullpath', 'name', 'object_type']
        cols.extend(keys)
        data = data[cols]

        return data

    # EXPORT--------------------------------------------------------------------
    def write_plots(self, fullpath):
        # type: (Union[str, Path]) -> RadonETL
        '''
        Writes metrics plots to given file.

        Args:
            fullpath (Path or str): Target file.

        Returns:
            RadonETL: self.
        '''
        def remove_test_modules(data):
            # type: (DataFrame) -> DataFrame
            mask = data.fullpath\
                .apply(lambda x: not re.search(r'_test\.py$', x)).astype(bool)
            return data[mask]

        lut = dict(
            h1='h1 - the number of distinct operators',
            h2='h2 - the number of distinct operands',
            n1='n1 - the total number of operators',
            n2='n2 - the total number of operands',
            vocabulary='vocabulary (h) - h1 + h2',
            length='length (N) - n1 + n2',
            calculated_length='calculated_length - h1 * log2(h1) + h2 * log2(h2)',
            volume='volume (V) - N * log2(h)',
            difficulty='difficulty (D) - h1 / 2 * n2 / h2',
            effort='effort (E) - D * V',
            time='time (T) - E / 18 seconds',
            bugs='bugs (B) - V / 3000 - an estimate of the errors in the implementation',
        )

        html = '<body style="background: #242424">\n'

        raw = remove_test_modules(self.raw_metrics)
        mi = remove_test_modules(self.maintainability_index)
        cc = remove_test_modules(self.cyclomatic_complexity_metrics)
        hal = remove_test_modules(self.halstead_metrics)

        raw['docstring_ratio'] = raw.multiline_comment / raw.code
        raw.sort_values('docstring_ratio', inplace=True)

        # line count
        fig = px.bar(
            raw,
            title='Line Count Metrics',
            x=raw.drop(columns='fullpath').columns.tolist(),
            y='fullpath',
            orientation='h',
            barmode='group',
            width=900,
            height=900,
            color_discrete_sequence=rpt.COLOR_SCALE,
        )
        fig.layout.update(rpt.PLOTLY_LAYOUT_THEME)
        html += fig.to_html()

        # maintainability
        fig = px.bar(
            mi,
            title='Maintainability Metrics',
            x='maintainability_index',
            y='fullpath',
            orientation='h',
            barmode='group',
            width=900,
            height=900,
            color_discrete_sequence=rpt.COLOR_SCALE,
        )
        fig.layout.update(rpt.PLOTLY_LAYOUT_THEME)
        html += fig.to_html()

        # cyclomatic
        fig = px.histogram(
            cc[['cyclomatic_complexity', 'cyclomatic_rank']],
            title='Cyclomatic Metric Distributions',
            nbins=10,
            width=900,
            height=500,
            color_discrete_sequence=rpt.COLOR_SCALE,
        )
        fig.layout.update(rpt.PLOTLY_LAYOUT_THEME)
        html += fig.to_html()

        # halstead
        cols = [
            'h1', 'h2', 'n1', 'n2', 'vocabulary', 'length', 'calculated_length',
            'volume', 'difficulty', 'effort', 'time', 'bugs'
        ]
        fig = px.histogram(
            hal[cols].rename(mapper=lambda x: lut[x], axis=1),
            title='Halstead Metric Distributions',
            nbins=10,
            width=1400,
            height=500,
            color_discrete_sequence=rpt.COLOR_SCALE,
        )
        fig.layout.update(rpt.PLOTLY_LAYOUT_THEME)
        html += fig.to_html()

        html += '\n</body>'

        with open(fullpath, 'w') as f:
            f.write(html)

        return self

    def write_tables(self, target_dir):
        # type: (Union[str, Path]) -> RadonETL
        '''
        Writes metrics tables as HTML files to given directory.

        Args:
            target_dir (Path or str): Target directory.

        Returns:
            RadonETL: self.
        '''
        def write_table(data, target):
            # type: (DataFrame, Path) -> None
            html = data.to_html()

            # make table sortable
            script = '<script '
            script += 'src="http://www.kryogenix.org/code/browser/sorttable/sorttable.js" '
            script += 'type="text/javascript"></script>\n'
            html = re.sub('class="dataframe"', 'class="sortable"', html)
            html = script + html

            with open(target, 'w') as f:
                f.write(html)

        data = self.data
        raw = self.raw_metrics
        mi = self.maintainability_index
        cc = self.cyclomatic_complexity_metrics
        hal = self.halstead_metrics

        write_table(data, Path(target_dir, 'all_metrics.html'))
        write_table(raw, Path(target_dir, 'raw_metrics.html'))
        write_table(mi, Path(target_dir, 'maintainability_metrics.html'))
        write_table(cc, Path(target_dir, 'cyclomatic_complexity_metrics.html'))
        write_table(hal, Path(target_dir, 'halstead_metrics.html'))

        return self



from typing import Dict, List  # noqa: F401

from pathlib import Path

from schematics import Model
from schematics.exceptions import ValidationError
from schematics.types import ListType, ModelType, StringType

Rules = List[Dict[str, str]]
# ------------------------------------------------------------------------------


def is_dir(dirpath):
    # type: (str) -> None
    '''
    Validates whether a given dirpath exists.

    Args:
        dirpath (str): Directory path.

    Raises:
        ValidationError: If dirpath is not a directory or does not exist.
    '''
    if not Path(dirpath).is_dir():
        msg = f'{dirpath} is not a directory or does not exist.'
        raise ValidationError(msg)


class ConformConfig(Model):
    '''
    A class for validating configurations supplied to ConformETL.

    Attributes:
        source_rules (Rules): A list of rules for parsing directories.
            Default: [].
        rename_rules (Rules): A list of rules for renaming source filepath
            to target filepaths. Default: [].
        group_rules (Rules): A list of rules for grouping files.
            Default: [].
        line_rules (Rules): A list of rules for peforming line copies and
            substitutions on files belonging to a given group. Default: [].
    '''
    class SourceRule(Model):
        path = StringType(required=True, validators=[is_dir])  # type: StringType
        include = StringType(required=False, serialize_when_none=False)  # type: StringType
        exclude = StringType(required=False, serialize_when_none=False)  # type: StringType

    class RenameRule(Model):
        regex = StringType(required=True)  # type: StringType
        replace = StringType(required=True)  # type: StringType

    class GroupRule(Model):
        name = StringType(required=True)  # type: StringType
        regex = StringType(required=True)  # type: StringType

    class LineRule(Model):
        group = StringType(required=True)  # type: StringType
        include = StringType(required=False, serialize_when_none=False)  # type: StringType
        exclude = StringType(required=False, serialize_when_none=False)  # type: StringType
        regex = StringType(required=False)  # type: StringType
        replace = StringType(required=False)  # type: StringType

    source_rules = ListType(ModelType(SourceRule), required=True)  # type: ListType
    rename_rules = ListType(ModelType(RenameRule), required=False)  # type: ListType
    group_rules = ListType(ModelType(GroupRule), required=False)  # type: ListType
    line_rules = ListType(ModelType(LineRule), required=False)  # type: ListType



'''
A library of generic tools for ETL work and visualization of JSON blobs and
python repositories.

Author: Alex Braun
Email: Alexander.G.Braun@gmail.com
Github: https://github.com/theNewFlesh
'''
from rolling_pin import blob_etl        # noqa F401
from rolling_pin import conform_config  # noqa F401
from rolling_pin import conform_etl     # noqa F401
from rolling_pin import repo_etl        # noqa F401
from rolling_pin import toml_etl        # noqa F401
from rolling_pin import tools           # noqa F401



from typing import Any, Dict, List, Union  # noqa: F401
from IPython.display import HTML, Image  # noqa: F401

from copy import deepcopy
from itertools import chain
from pathlib import Path
import re

from lunchbox.enforce import Enforce
from pandas import DataFrame
import lunchbox.tools as lbt
import yaml

from rolling_pin.blob_etl import BlobETL
from rolling_pin.conform_config import ConformConfig
import rolling_pin.tools as rpt

Rules = List[Dict[str, str]]
# ------------------------------------------------------------------------------


CONFORM_COLOR_SCHEME = deepcopy(rpt.COLOR_SCHEME)
CONFORM_COLOR_SCHEME.update({
    'node_font': '#DE958E',
    'node_value_font': '#B6ECF3',
    'edge': '#DE958E',
    'edge_value': '#B6ECF3',
    'node_library_font': '#B6ECF3',
    'node_module_font': '#DE958E',
    'edge_library': '#B6ECF3',
    'edge_module': '#DE958E'
})


class ConformETL:
    '''
    ConformETL creates a DataFrame from a given directory of source files.
    Then it generates target paths given a set of rules.
    Finally, the conform method is called and the source files are copied to
    their target filepaths.
    '''
    @staticmethod
    def _get_data(
        source_rules=[], rename_rules=[], group_rules=[], line_rules=[]
    ):
        # type: (Rules, Rules, Rules, Rules) -> DataFrame
        '''
        Generates DataFrame from given source_rules and then generates target
        paths for them given other rules.

        Args:
            source_rules (Rules): A list of rules for parsing directories.
                 Default: [].
            rename_rules (Rules): A list of rules for renaming source filepath
                to target filepaths. Default: [].
            group_rules (Rules): A list of rules for grouping files.
                Default: [].
            line_rules (Rules): A list of rules for peforming line copies on
                files belonging to a given group. Default: [].

        Returns:
            DataFrame: Conform DataFrame.
        '''
        # source
        source = []  # type: List[Any]
        for rule in source_rules:
            files = rpt.list_all_files(
                rule['path'],
                include_regex=rule.get('include', None),
                exclude_regex=rule.get('exclude', None),
            )
            source.extend(files)
        source = sorted([x.as_posix() for x in source])
        data = DataFrame()
        data['source'] = source
        data['target'] = source

        # rename
        for rule in rename_rules:
            data.target = data.target.apply(
                lambda x: rpt.replace_and_format(
                    rule['regex'], rule['replace'], x
                )
            )

        # group
        data['groups'] = data.source.apply(lambda x: [])
        for rule in group_rules:
            mask = data.source \
                .apply(lambda x: re.search(rule['regex'], x)) \
                .astype(bool)
            data.loc[mask, 'groups'] = data.groups \
                .apply(lambda x: x + [rule['name']])
        mask = data.groups.apply(lambda x: x == [])
        data.loc[mask, 'groups'] = data.loc[mask, 'groups'] \
            .apply(lambda x: ['base'])

        # line
        groups = set([x['group'] for x in line_rules])
        data['line_rule'] = data.groups \
            .apply(lambda x: len(set(x).intersection(groups)) > 0)

        return data

    @classmethod
    def from_yaml(cls, filepath):
        # type: (Union[str, Path]) -> ConformETL
        '''
        Construct ConformETL instance from given yaml file.

        Args:
            filepath (str or Path): YAML file.

        Raises:
            EnforceError: If file does not end in yml or yaml.

        Returns:
            ConformETL: ConformETL instance.
        '''
        filepath = Path(filepath).as_posix()
        ext = Path(filepath).suffix[1:].lower()
        msg = f'{filepath} does not end in yml or yaml.'
        Enforce(ext, 'in', ['yml', 'yaml'], message=msg)
        # ----------------------------------------------------------------------

        with open(filepath) as f:
            config = yaml.safe_load(f)
        return cls(**config)

    def __init__(
        self, source_rules=[], rename_rules=[], group_rules=[], line_rules=[]
    ):
        # type: (Rules, Rules, Rules, Rules) -> None
        '''
        Generates DataFrame from given source_rules and then generates target
        paths for them given other rules.

        Args:
            source_rules (Rules): A list of rules for parsing directories.
                 Default: [].
            rename_rules (Rules): A list of rules for renaming source filepath
                to target filepaths. Default: [].
            group_rules (Rules): A list of rules for grouping files.
                Default: [].
            line_rules (Rules): A list of rules for peforming line copies on
                files belonging to a given group. Default: [].

        Raises:
            DataError: If configuration is invalid.
        '''
        config = dict(
            source_rules=source_rules,
            rename_rules=rename_rules,
            group_rules=group_rules,
            line_rules=line_rules,
        )
        cfg = ConformConfig(config)
        cfg.validate()
        config = cfg.to_native()

        self._data = self._get_data(
            source_rules=source_rules,
            rename_rules=rename_rules,
            group_rules=group_rules,
            line_rules=line_rules,
        )  # type: DataFrame
        self._line_rules = line_rules  # type: Rules

    def __repr__(self):
        # type: () -> str
        '''
        String representation of conform DataFrame.

        Returns:
            str: Table optimized for output to shell.
        '''
        data = self._data.copy()
        data.line_rule = data.line_rule.apply(lambda x: 'X' if x else '')
        data.rename(lambda x: x.upper(), axis=1, inplace=True)
        output = data \
            .to_string(index=False, max_colwidth=150, col_space=[50, 50, 20, 10])
        return output

    @property
    def groups(self):
        # type: () -> List[str]
        '''
        list[str]: List of groups found with self._data.
        '''
        output = self._data.groups.tolist()
        output = sorted(list(set(chain(*output))))
        output.remove('base')
        output.insert(0, 'base')
        return output

    def to_dataframe(self):
        # type: () -> DataFrame
        '''
        Returns:
            DataFrame: Copy of internal data.
        '''
        return self._data.copy()

    def to_blob(self):
        # type: () -> BlobETL
        '''
        Converts self into a BlobETL object with target column as keys and
        source columns as values.

        Returns:
            BlobETL: BlobETL of target and source filepaths.
        '''
        data = self._data
        keys = data.target.tolist()
        vals = data.source.tolist()
        output = dict(zip(keys, vals))
        return BlobETL(output)

    def to_html(
        self, orient='lr', color_scheme=CONFORM_COLOR_SCHEME, as_png=False
    ):
        # type: (str, Dict[str, str], bool) -> Union[Image, HTML]
        '''
        For use in inline rendering of graph data in Jupyter Lab.
        Graph from target to source filepath. Target is in red, source is in
        cyan.

        Args:
            orient (str, optional): Graph layout orientation. Default: lr.
                Options include:

                * tb - top to bottom
                * bt - bottom to top
                * lr - left to right
                * rl - right to left
            color_scheme: (dict, optional): Color scheme to be applied to graph.
                Default: rolling_pin.conform_etl.CONFORM_COLOR_SCHEME
            as_png (bool, optional): Display graph as a PNG image instead of
                SVG. Useful for display on Github. Default: False.

        Returns:
            IPython.display.HTML: HTML object for inline display.
        '''
        return self.to_blob() \
            .to_html(orient=orient, color_scheme=color_scheme, as_png=as_png)

    def conform(self, groups='all'):
        # type: (Union[str, List[str]]) -> None
        '''
        Copies source files to target filepaths.

        Args:
            groups (str or list[str]): Groups of files which are to be conformed.
                'all' means all groups. Default: 'all'.
        '''
        if isinstance(groups, str):
            groups = [groups]
        if groups == ['all']:
            groups = self.groups

        data = self.to_dataframe()

        # copy files
        grps = set(groups)
        mask = data.groups \
            .apply(lambda x: set(x).intersection(grps)) \
            .apply(lambda x: len(x) > 0)
        data = data[mask]
        data.apply(lambda x: rpt.copy_file(x.source, x.target), axis=1)

        # copy lines
        data['text'] = data.source.apply(lambda x: lbt.try_(rpt.read_text, x, 'error'))
        readable_mask = data.text.apply(lambda x: isinstance(x, str))
        data.loc[~readable_mask, 'text'] = ''
        rules = list(filter(lambda x: x['group'] in groups, self._line_rules))
        for rule in rules:
            mask = data.groups.apply(lambda x: rule['group'] in x)
            data.loc[mask, 'text'] = data.loc[mask, 'text'].apply(
                lambda x: rpt.filter_text(
                    x,
                    include_regex=rule.get('include', None),
                    exclude_regex=rule.get('exclude', None),
                    replace_regex=rule.get('regex', None),
                    replace_value=rule.get('replace', None),
                )
            )
        data[readable_mask].apply(lambda x: rpt.write_text(x.text, x.target), axis=1)



from typing import Any, Dict, Generator, Iterable, List, Optional, Union  # noqa: F401
import pydot  # noqa: F401

from collections import OrderedDict
from pathlib import Path
import logging
import os
import re
import shutil

from IPython.display import HTML, Image
import pandas as pd

Filepath = Union[str, Path]
LOG_LEVEL = os.environ.get('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=LOG_LEVEL)
LOGGER = logging.getLogger(__name__)
# ------------------------------------------------------------------------------

'''
Contains basic functions for more complex ETL functions and classes.
'''

# COLOR-SCHEME------------------------------------------------------------------
COLOR_SCHEME = dict(
    background='#242424',
    node='#343434',
    node_font='#B6ECF3',
    node_value='#343434',
    node_value_font='#DE958E',
    edge='#B6ECF3',
    edge_value='#DE958E',
    node_library_font='#DE958E',
    node_subpackage_font='#A0D17B',
    node_module_font='#B6ECF3',
    edge_library='#DE958E',
    edge_subpackage='#A0D17B',
    edge_module='#B6ECF3',
)  # type: Dict[str, str]


PLOTLY_COLOR_SCHEME = dict(
    bg='#242424',
    blue1='#5F95DE',
    blue2='#93B6E6',
    cyan1='#7EC4CF',
    cyan2='#B6ECF3',
    dark1='#040404',
    dark2='#141414',
    dialog1='#444459',
    dialog2='#5D5D7A',
    green1='#8BD155',
    green2='#A0D17B',
    grey1='#343434',
    grey2='#444444',
    light1='#A4A4A4',
    light2='#F4F4F4',
    orange1='#EB9E58',
    orange2='#EBB483',
    purple1='#C98FDE',
    purple2='#AC92DE',
    red1='#F77E70',
    red2='#DE958E',
    yellow1='#E8EA7E',
    yellow2='#E9EABE',
)


COLOR_SCALE = [
    'cyan2', 'red2', 'green2', 'blue2', 'orange2', 'purple2', 'yellow2',
    'light2', 'cyan1', 'red1', 'green1', 'blue1'
]  # type: List[str]
COLOR_SCALE = [PLOTLY_COLOR_SCHEME[x] for x in COLOR_SCALE]


PLOTLY_LAYOUT_THEME = {
    'legend': {
        'bgcolor': PLOTLY_COLOR_SCHEME['bg'],
        'title': {'text': ''},
        'font': {
            'color': PLOTLY_COLOR_SCHEME['light2']
        }
    },
    'paper_bgcolor': PLOTLY_COLOR_SCHEME['bg'],
    'plot_bgcolor': PLOTLY_COLOR_SCHEME['bg'],
    'title': {
        'font': {
            'color': PLOTLY_COLOR_SCHEME['light2']
        }
    },
    'xaxis': {
        'gridcolor': PLOTLY_COLOR_SCHEME['grey1'],
        'showgrid': True,
        'tickfont': {
            'color': PLOTLY_COLOR_SCHEME['light1']
        },
        'title': {
            'font': {
                'color': PLOTLY_COLOR_SCHEME['light1']
            }
        },
        'zerolinecolor': PLOTLY_COLOR_SCHEME['grey2']
    },
    'yaxis': {
        'gridcolor': PLOTLY_COLOR_SCHEME['grey1'],
        'showgrid': True,
        'tickfont': {
            'color': PLOTLY_COLOR_SCHEME['light1']
        },
        'title': {
            'font': {
                'color': PLOTLY_COLOR_SCHEME['light1']
            }
        },
        'zerolinecolor': PLOTLY_COLOR_SCHEME['grey2']
    }
}


# PREDICATE-FUNCTIONS-----------------------------------------------------------
def is_iterable(item):
    # type: (Any) -> bool
    '''
    Determines if given item is iterable.

    Args:
        item (object): Object to be tested.

    Returns:
        bool: Whether given item is iterable.
    '''
    if is_listlike(item) or is_dictlike(item):
        return True
    return False


def is_dictlike(item):
    # type: (Any) -> bool
    '''
    Determines if given item is dict-like.

    Args:
        item (object): Object to be tested.

    Returns:
        bool: Whether given item is dict-like.
    '''
    for type_ in [dict, OrderedDict]:
        if isinstance(item, type_):
            if item.__class__.__name__ == 'Counter':
                return False
            return True
    return False


def is_listlike(item):
    # type: (Any) -> bool
    '''
    Determines if given item is list-like.

    Args:
        item (object): Object to be tested.

    Returns:
        bool: Whether given item is list-like.
    '''
    for type_ in [list, tuple, set]:
        if isinstance(item, type_):
            return True
    return False


# CORE-FUNCTIONS----------------------------------------------------------------
def flatten(item, separator='/', embed_types=True):
    # type: (Iterable, str, bool) -> Dict[str, Any]
    '''
    Flattens a iterable object into a flat dictionary.

    Args:
        item (object): Iterable object.
        separator (str, optional): Field separator in keys. Default: '/'.

    Returns:
        dict: Dictionary representation of given object.
    '''
    output = {}  # type: Dict[str, Any]

    def recurse(item, cursor):
        # type (Iterable, Any) -> None
        if is_listlike(item):
            if embed_types:
                name = item.__class__.__name__
                item = [(f'<{name}_{i}>', val) for i, val in enumerate(item)]
                item = dict(item)
            else:
                item = dict(enumerate(item))
        if is_dictlike(item):
            for key, val in item.items():
                new_key = f'{cursor}{separator}{str(key)}'
                if is_iterable(val) and len(val) > 0:
                    recurse(val, new_key)
                else:
                    final_key = re.sub('^' + separator, '', new_key)
                    output[final_key] = val

    recurse(item, '')
    return output


def nest(flat_dict, separator='/'):
    # type: (Dict[str, Any], str) -> Dict[str, Any]
    '''
    Converts a flat dictionary into a nested dictionary by splitting keys by a
    given separator.

    Args:
        flat_dict (dict): Flat dictionary.
        separator (str, optional): Field separator within given dictionary's
            keys. Default: '/'.

    Returns:
        dict: Nested dictionary.
    '''
    output = {}  # type: Dict[str, Any]
    for keys, val in flat_dict.items():
        split_keys = list(filter(
            lambda x: x != '', keys.split(separator)
        ))
        cursor = output
        last = split_keys.pop()
        for key in split_keys:
            if key not in cursor:
                cursor[key] = {}

            if not isinstance(cursor[key], dict):
                msg = f"Duplicate key conflict. Key: '{key}'."
                raise KeyError(msg)

            cursor = cursor[key]
        cursor[last] = val
    return output


def unembed(item):
    # type: (Any) -> Any
    '''
    Convert embeded types in dictionary keys into python types.

    Args:
        item (object): Dictionary with embedded types.

    Returns:
        object: Converted object.
    '''
    lut = {'list': list, 'tuple': tuple, 'set': set}
    embed_re = re.compile(r'^<([a-z]+)_(\d+)>$')

    if is_dictlike(item) and item != {}:
        output = {}  # type: Any
        keys = list(item.keys())
        match = embed_re.match(keys[0])
        if match:
            indices = [embed_re.match(key).group(2) for key in keys]  # type: ignore
            indices = map(int, indices)  # type: ignore

            output = []
            for i, key in sorted(zip(indices, keys)):
                next_item = item[key]
                if is_dictlike(next_item):
                    next_item = unembed(next_item)
                output.append(next_item)

            output = lut[match.group(1)](output)
            return output
        else:
            for key, val in item.items():
                output[key] = unembed(val)
        return output
    return item


# FILE-FUNCTIONS----------------------------------------------------------------
def list_all_files(
    directory,           # type: Filepath
    include_regex=None,  # type: Optional[str]
    exclude_regex=None   # type: Optional[str]
):
    # type: (...) -> Generator[Path, None, None]
    '''
    Recusively list all files within a given directory.

    Args:
        directory (str or Path): Directory to walk.
        include_regex (str, optional): Include filenames that match this regex.
            Default: None.
        exclude_regex (str, optional): Exclude filenames that match this regex.
            Default: None.

    Raises:
        FileNotFoundError: If argument is not a directory or does not exist.

    Yields:
        Path: File.
    '''
    directory = Path(directory)
    if not directory.is_dir():
        msg = f'{directory} is not a directory or does not exist.'
        raise FileNotFoundError(msg)

    include_re = re.compile(include_regex or '')  # type: Any
    exclude_re = re.compile(exclude_regex or '')  # type: Any

    for root, _, files in os.walk(directory):
        for file_ in files:
            filepath = Path(root, file_)

            output = True
            temp = filepath.absolute().as_posix()
            if include_regex is not None and not include_re.search(temp):
                output = False
            if exclude_regex is not None and exclude_re.search(temp):
                output = False

            if output:
                yield Path(root, file_)


def directory_to_dataframe(directory, include_regex='', exclude_regex=r'\.DS_Store'):
    # type: (Filepath, str, str) -> pd.DataFrame
    r'''
    Recursively list files with in a given directory as rows in a pd.DataFrame.

    Args:
        directory (str or Path): Directory to walk.
        include_regex (str, optional): Include filenames that match this regex.
            Default: None.
        exclude_regex (str, optional): Exclude filenames that match this regex.
            Default: '\.DS_Store'.

    Returns:
        pd.DataFrame: pd.DataFrame with one file per row.
    '''
    files = list_all_files(
        directory,
        include_regex=include_regex,
        exclude_regex=exclude_regex
    )  # type: Any
    files = sorted(list(files))

    data = pd.DataFrame()
    data['filepath'] = files
    data['filename'] = data.filepath.apply(lambda x: x.name)
    data['extension'] = data.filepath \
        .apply(lambda x: Path(x).suffix.lstrip('.'))
    data.filepath = data.filepath.apply(lambda x: x.absolute().as_posix())
    return data


def get_parent_fields(key, separator='/'):
    # type: (str, str) -> List[str]
    '''
    Get all the parent fields of a given key, split by given separator.

    Args:
        key (str): Key.
        separator (str, optional): String that splits key into fields.
            Default: '/'.

    Returns:
        list(str): List of absolute parent fields.
    '''
    fields = key.split(separator)
    output = []  # type: List[str]
    for i in range(len(fields) - 1):
        output.append(separator.join(fields[:i + 1]))
    return output


def filter_text(
    text,                # type: str
    include_regex=None,  # type: Optional[str]
    exclude_regex=None,  # type: Optional[str]
    replace_regex=None,  # type: Optional[str]
    replace_value=None,  # type: Optional[str]
):
    # type: (...) -> str
    '''
    Filter given text by applying regular expressions to each line.

    Args:
        text (str): Newline separated lines.
        include_regex (str, optional): Keep lines that match given regex.
            Default: None.
        exclude_regex (str, optional): Remove lines that match given regex.
            Default: None.
        replace_regex (str, optional): Substitutes regex matches in lines with
            replace_value. Default: None.
        replace_value (str, optional): Regex substitution value. Default: ''.

    Raises:
        AssertionError: If source is not a file.

    Returns:
        str: Filtered text.
    '''
    lines = text.split('\n')
    if include_regex is not None:
        lines = list(filter(lambda x: re.search(include_regex, x), lines))
    if exclude_regex is not None:
        lines = list(filter(lambda x: not re.search(exclude_regex, x), lines))
    if replace_regex is not None:
        rep_val = replace_value or ''
        lines = [re.sub(replace_regex, rep_val, x) for x in lines]
    output = '\n'.join(lines)
    return output


def read_text(filepath):
    # type: (Filepath) -> str
    '''
    Convenience function for reading text from given file.

    Args:
        filepath (str or Path): File to be read.

    Raises:
        AssertionError: If source is not a file.

    Returns:
        str: text.
    '''
    assert Path(filepath).is_file()
    with open(filepath) as f:
        return f.read()


def write_text(text, filepath):
    # type: (str, Filepath) -> None
    '''
    Convenience function for writing text to given file.
    Creates directories as needed.

    Args:
        text (str): Text to be written.
        filepath (str or Path): File to be written.
    '''
    os.makedirs(Path(filepath).parent, exist_ok=True)
    with open(filepath, 'w') as f:
        f.write(text)


def copy_file(source, target):
    # type: (Filepath, Filepath) -> None
    '''
    Copy a source file to a target file. Creating directories as needed.

    Args:
        source (str or Path): Source filepath.
        target (str or Path): Target filepath.

    Raises:
        AssertionError: If source is not a file.
    '''
    assert Path(source).is_file()
    os.makedirs(Path(target).parent, exist_ok=True)
    shutil.copy2(source, target)


def move_file(source, target):
    # type: (Filepath, Filepath) -> None
    '''
    Moves a source file to a target file. Creating directories as needed.

    Args:
        source (str or Path): Source filepath.
        target (str or Path): Target filepath.

    Raises:
        AssertionError: If source is not a file.
    '''
    src = Path(source).as_posix()
    assert Path(src).is_file()
    os.makedirs(Path(target).parent, exist_ok=True)
    shutil.move(src, target)


# EXPORT-FUNCTIONS--------------------------------------------------------------
def dot_to_html(dot, layout='dot', as_png=False):
    # type: (pydot.Dot, str, bool) -> Union[HTML, Image]
    '''
    Converts a given pydot graph into a IPython.display.HTML object.
    Used in jupyter lab inline display of graph data.

    Args:
        dot (pydot.Dot): Pydot Graph instance.
        layout (str, optional): Graph layout style.
            Options include: circo, dot, fdp, neato, sfdp, twopi.
            Default: dot.
        as_png (bool, optional): Display graph as a PNG image instead of SVG.
            Useful for display on Github. Default: False.

    Raises:
        ValueError: If invalid layout given.

    Returns:
        IPython.display.HTML: HTML instance.
    '''
    layouts = ['circo', 'dot', 'fdp', 'neato', 'sfdp', 'twopi']
    if layout not in layouts:
        msg = f'Invalid layout value. {layout} not in {layouts}.'
        raise ValueError(msg)

    if as_png:
        return Image(data=dot.create_png())

    svg = dot.create_svg(prog=layout)
    html = f'<object type="image/svg+xml" data="data:image/svg+xml;{svg}"></object>'  # type: Any
    html = HTML(html)
    html.data = re.sub(r'\\n|\\', '', html.data)
    html.data = re.sub('</svg>.*', '</svg>', html.data)
    return html


def write_dot_graph(
    dot,
    fullpath,
    layout='dot',
):
    # type: (pydot.Dot, Union[str, Path], str) -> None
    '''
    Writes a pydot.Dot object to a given filepath.
    Formats supported: svg, dot, png.

    Args:
        dot (pydot.Dot): Pydot Dot instance.
        fulllpath (str or Path): File to be written to.
        layout (str, optional): Graph layout style.
            Options include: circo, dot, fdp, neato, sfdp, twopi. Default: dot.

    Raises:
        ValueError: If invalid file extension given.
    '''
    if isinstance(fullpath, Path):
        fullpath = Path(fullpath).absolute().as_posix()

    _, ext = os.path.splitext(fullpath)
    ext = re.sub(r'^\.', '', ext)
    if re.search('^svg$', ext, re.I):
        dot.write_svg(fullpath, prog=layout)
    elif re.search('^dot$', ext, re.I):
        dot.write_dot(fullpath, prog=layout)
    elif re.search('^png$', ext, re.I):
        dot.write_png(fullpath, prog=layout)
    else:
        msg = f'Invalid extension found: {ext}. '
        msg += 'Valid extensions include: svg, dot, png.'
        raise ValueError(msg)


# MISC-FUNCTIONS----------------------------------------------------------------
def replace_and_format(regex, replace, string, flags=0):
    # type: (str, str, str, Any) -> str
    r'''
    Perform a regex substitution on a given string and format any named group
    found in the result with groupdict data from the pattern. Group beggining
    with 'i' will be converted to integers. Groups beggining with 'f' will be
    converted to floats.

    ----------------------------------------------------------------------------

    Named group anatomy:
    ====================
        * (?P<NAME>PATTERN)
        * NAME becomes a key and whatever matches PATTERN becomes its value.
        >>> re.search('(?P<i>\d+)', 'foobar123').groupdict()
        {'i': '123'}

    ----------------------------------------------------------------------------

    Examples:
    =========
        Special groups:
            * (?P<i>\d)     - string matched by '\d' will be converted to an integer
            * (?P<f>\d)     - string matched by '\d' will be converted to an float
            * (?P<i_foo>\d) - string matched by '\d' will be converted to an integer
            * (?P<f_bar>\d) - string matched by '\d' will be converted to an float

        Named groups (long):
            >>> proj = '(?P<p>[a-z0-9]+)'
            >>> spec = '(?P<s>[a-z0-9]+)'
            >>> desc = '(?P<d>[a-z0-9\-]+)'
            >>> ver = '(?P<iv>\d+)\.'
            >>> frame = '(?P<i_f>\d+)'
            >>> regex = f'{proj}\.{spec}\.{desc}\.v{ver}\.{frame}.*'
            >>> replace = 'p-{p}_s-{s}_d-{d}_v{iv:03d}_f{i_f:04d}.jpeg'
            >>> string = 'proj.spec.desc.v1.25.png'
            >>> replace_and_format(regex, replace, string, flags=re.IGNORECASE)
            p-proj_s-spec_d-desc_v001_f0025.jpeg

        Named groups (short):
            >>> replace_and_format(
                '(?P<p>[a-z0-9]+)\.(?P<s>[a-z0-9]+)\.(?P<d>[a-z0-9\-]+)\.v(?P<iv>\d+)\.(?P<i_f>\d+).*',
                'p-{p}_s-{s}_d-{d}_v{iv:03d}_f{i_f:04d}.jpeg',
                'proj.spec.desc.v1.25.png',
            )
            p-proj_s-spec_d-desc_v001_f0025.jpeg

        No groups:
            >>> replace_and_format('foo', 'bar', 'foobar')
            barbar

    ----------------------------------------------------------------------------

    Args:
        regex (str): Regex pattern to search string with.
        replace (str): Replacement string which may contain formart variables
            ie '{variable}'.
        string (str): String to be converted.
        flags (object, optional): re.sub flags. Default: 0.

    Returns:
        str: Converted string.
    '''
    match = re.search(regex, string, flags=flags)
    grp = {}
    if match:
        grp = match.groupdict()

    for key, val in grp.items():
        if key.startswith('f'):
            grp[key] = float(val)
        elif key.startswith('i'):
            grp[key] = int(val)

    output = re.sub(regex, replace, string, flags=flags)
    # .format won't evaluate math expressions so do this
    if grp != {}:
        output = eval(f"f'{output}'", None, grp)
    return output



from typing import Any, Type, TypeVar, Union  # noqa: F401

from copy import deepcopy
from pathlib import Path
import os

from lunchbox.enforce import Enforce
import toml

from rolling_pin.blob_etl import BlobETL
from toml.decoder import TomlDecodeError

T = TypeVar('T', bound='TomlETL')
# ------------------------------------------------------------------------------


class TomlETL:
    @classmethod
    def from_string(cls, text):
        # type: (Type[T], str) -> T
        '''
        Creates a TomlETL instance from a given TOML string.

        Args:
            text (str): TOML string.

        Returns:
            TomlETL: TomlETL instance.
        '''
        return cls(toml.loads(text))

    @classmethod
    def from_toml(cls, filepath):
        # type: (Type[T], Union[str, Path]) -> T
        '''
        Creates a TomlETL instance from a given TOML file.

        Args:
            filepath (str or Path): TOML file.

        Returns:
            TomlETL: TomlETL instance.
        '''
        return cls(toml.load(filepath))

    def __init__(self, data):
        # type: (dict[str, Any]) -> None
        '''
        Creates a TomlETL instance from a given dictionary.

        Args:
            data (dict): Dictionary.
        '''
        self._data = data

    def to_dict(self):
        # type: () -> dict
        '''
        Converts instance to dictionary copy.

        Returns:
            dict: Dictionary copy of instance.
        '''
        return deepcopy(self._data)

    def to_string(self):
        # type: () -> str
        '''
        Converts instance to a TOML formatted string.

        Returns:
            str: TOML string.
        '''
        return toml.dumps(
            self._data, encoder=toml.TomlArraySeparatorEncoder(separator=',')
        )

    def write(self, filepath):
        # type: (Union[str, Path]) -> None
        '''
        Writes instance to given TOML file.

        Args:
            filepath (str or Path): Target filepath.
        '''
        filepath = Path(filepath)
        os.makedirs(filepath.parent, exist_ok=True)
        with open(filepath, 'w') as f:
            toml.dump(
                self._data,
                f,
                encoder=toml.TomlArraySeparatorEncoder(separator=',')
            )

    def edit(self, patch):
        # type: (str) -> TomlETL
        '''
        Apply edit to internal data given TOML patch.
        Patch is always of the form '[key]=[value]' and in TOML format.

        Args:
            patch (str): TOML patch to be applied.

        Raises:
            TOMLDecoderError: If patch cannot be decoded.
            EnforceError: If '=' not found in patch.

        Returns:
            TomlETL: New TomlETL instance with edits.
        '''
        msg = 'Edit patch must be a TOML parsable key value snippet with a "=" '
        msg += 'character.'
        try:
            toml.loads(patch)
        except TomlDecodeError as e:
            msg += ' ' + e.msg
            raise TomlDecodeError(msg, e.doc, e.pos)
        Enforce('=', 'in', patch, message=msg)
        # ----------------------------------------------------------------------

        key, val = patch.split('=', maxsplit=1)
        val = toml.loads(f'x={val}')['x']
        data = BlobETL(self._data, separator='.').to_flat_dict()
        data[key] = val
        data = BlobETL(data, separator='.').to_dict()
        return TomlETL(data)

    def delete(self, regex):
        # type: (str) -> TomlETL
        '''
        Returns portion of data whose keys fo not match a given regular expression.

        Args:
            regex (str): Regular expression applied to keys.

        Returns:
            TomlETL: New TomlETL instance.
        '''
        data = BlobETL(self._data, separator='.') \
            .query(regex, ignore_case=False, invert=True) \
            .to_dict()
        return TomlETL(data)

    def search(self, regex):
        # type: (str) -> TomlETL
        '''
        Returns portion of data whose keys match a given regular expression.

        Args:
            regex (str): Regular expression applied to keys.

        Returns:
            TomlETL: New TomlETL instance.
        '''
        data = BlobETL(self._data, separator='.') \
            .query(regex, ignore_case=False) \
            .to_dict()
        return TomlETL(data)
